@@ -12,7 +12,7 @@ import os
 import copy
 
 from mercurial import hg, util, cmdutil, scmutil, match as match_, \
-        archival, pathutil, revset, error
+        archival, pathutil, registrar, revset, error
 from mercurial.i18n import _
 
 import lfutil
@@ -141,8 +141,7 @@ def addlargefiles(ui, repo, isaddremove, matcher, **opts):
 
     # Need to lock, otherwise there could be a race condition between
     # when standins are created and added to the repo.
-    wlock = repo.wlock()
-    try:
+    with repo.wlock():
         if not opts.get('dry_run'):
             standins = []
             lfdirstate = lfutil.openlfdirstate(ui, repo)
@@ -161,8 +160,6 @@ def addlargefiles(ui, repo, isaddremove, matcher, **opts):
                     if f in m.files()]
 
         added = [f for f in lfnames if f not in bad]
-    finally:
-        wlock.release()
     return added, bad
 
 def removelargefiles(ui, repo, isaddremove, matcher, **opts):
@@ -199,8 +196,7 @@ def removelargefiles(ui, repo, isaddremove, matcher, **opts):
 
     # Need to lock because standin files are deleted then removed from the
     # repository and we could race in-between.
-    wlock = repo.wlock()
-    try:
+    with repo.wlock():
         lfdirstate = lfutil.openlfdirstate(ui, repo)
         for f in sorted(remove):
             if ui.verbose or not m.exact(f):
@@ -231,8 +227,6 @@ def removelargefiles(ui, repo, isaddremove, matcher, **opts):
                                   False)
 
         lfdirstate.write()
-    finally:
-        wlock.release()
 
     return result
 
@@ -458,11 +452,10 @@ def overridecheckunknownfile(origfn, repo, wctx, mctx, f, f2=None):
 # writing the files into the working copy and lfcommands.updatelfiles
 # will update the largefiles.
 def overridecalculateupdates(origfn, repo, p1, p2, pas, branchmerge, force,
-                             partial, acceptremote, followcopies):
+                             acceptremote, *args, **kwargs):
     overwrite = force and not branchmerge
     actions, diverge, renamedelete = origfn(
-        repo, p1, p2, pas, branchmerge, force, partial, acceptremote,
-        followcopies)
+        repo, p1, p2, pas, branchmerge, force, acceptremote, *args, **kwargs)
 
     if overwrite:
         return actions, diverge, renamedelete
@@ -470,17 +463,20 @@ def overridecalculateupdates(origfn, repo, p1, p2, pas, branchmerge, force,
     # Convert to dictionary with filename as key and action as value.
     lfiles = set()
     for f in actions:
-        splitstandin = f and lfutil.splitstandin(f)
+        splitstandin = lfutil.splitstandin(f)
         if splitstandin in p1:
             lfiles.add(splitstandin)
         elif lfutil.standin(f) in p1:
             lfiles.add(f)
 
-    for lfile in lfiles:
+    for lfile in sorted(lfiles):
         standin = lfutil.standin(lfile)
         (lm, largs, lmsg) = actions.get(lfile, (None, None, None))
         (sm, sargs, smsg) = actions.get(standin, (None, None, None))
         if sm in ('g', 'dc') and lm != 'r':
+            if sm == 'dc':
+                f1, f2, fa, move, anc = sargs
+                sargs = (p2[f2].flags(), False)
             # Case 1: normal file in the working copy, largefile in
             # the second parent
             usermsg = _('remote turned local normal file %s into a largefile\n'
@@ -496,6 +492,9 @@ def overridecalculateupdates(origfn, repo, p1, p2, pas, branchmerge, force,
                 else:
                     actions[standin] = ('r', None, 'replaced by non-standin')
         elif lm in ('g', 'dc') and sm != 'r':
+            if lm == 'dc':
+                f1, f2, fa, move, anc = largs
+                largs = (p2[f2].flags(), False)
             # Case 2: largefile in the working copy, normal file in
             # the second parent
             usermsg = _('remote turned local largefile %s into a normal file\n'
@@ -538,7 +537,7 @@ def mergerecordupdates(orig, repo, actions, branchmerge):
 # largefiles. This will handle identical edits without prompting the user.
 def overridefilemerge(origfn, premerge, repo, mynode, orig, fcd, fco, fca,
                       labels=None):
-    if not lfutil.isstandin(orig):
+    if not lfutil.isstandin(orig) or fcd.isabsent() or fco.isabsent():
         return origfn(premerge, repo, mynode, orig, fcd, fco, fca,
                       labels=labels)
 
@@ -555,7 +554,7 @@ def overridefilemerge(origfn, premerge, repo, mynode, orig, fcd, fco, fca,
                (lfutil.splitstandin(orig), ahash, dhash, ohash),
              0) == 1)):
         repo.wwrite(fcd.path(), fco.data(), fco.flags())
-    return True, 0
+    return True, 0, False
 
 def copiespathcopies(orig, ctx1, ctx2, match=None):
     copies = orig(ctx1, ctx2, match=match)
@@ -604,7 +603,7 @@ def overridecopy(orig, ui, repo, pats, opts, rename=False):
 
     def makestandin(relpath):
         path = pathutil.canonpath(repo.root, repo.getcwd(), relpath)
-        return os.path.join(repo.wjoin(lfutil.standin(path)))
+        return repo.wvfs.join(lfutil.standin(path))
 
     fullpats = scmutil.expandpats(pats)
     dest = fullpats[-1]
@@ -674,7 +673,7 @@ def overridecopy(orig, ui, repo, pats, opts, rename=False):
                 dest.startswith(repo.wjoin(lfutil.shortname))):
                 srclfile = src.replace(repo.wjoin(lfutil.standin('')), '')
                 destlfile = dest.replace(repo.wjoin(lfutil.standin('')), '')
-                destlfiledir = os.path.dirname(repo.wjoin(destlfile)) or '.'
+                destlfiledir = repo.wvfs.dirname(repo.wjoin(destlfile)) or '.'
                 if not os.path.isdir(destlfiledir):
                     os.makedirs(destlfiledir)
                 if rename:
@@ -717,16 +716,15 @@ def overriderevert(orig, ui, repo, ctx, parents, *pats, **opts):
     # Because we put the standins in a bad state (by updating them)
     # and then return them to a correct state we need to lock to
     # prevent others from changing them in their incorrect state.
-    wlock = repo.wlock()
-    try:
+    with repo.wlock():
         lfdirstate = lfutil.openlfdirstate(ui, repo)
         s = lfutil.lfdirstatestatus(lfdirstate, repo)
         lfdirstate.write()
         for lfile in s.modified:
             lfutil.updatestandin(repo, lfutil.standin(lfile))
         for lfile in s.deleted:
-            if (os.path.exists(repo.wjoin(lfutil.standin(lfile)))):
-                os.unlink(repo.wjoin(lfutil.standin(lfile)))
+            if (repo.wvfs.exists(lfutil.standin(lfile))):
+                repo.wvfs.unlink(lfutil.standin(lfile))
 
         oldstandins = lfutil.getstandinsstate(repo)
 
@@ -778,9 +776,6 @@ def overriderevert(orig, ui, repo, ctx, parents, *pats, **opts):
         lfcommands.updatelfiles(ui, repo, filelist, printmessage=False,
                                 normallookup=True)
 
-    finally:
-        wlock.release()
-
 # after pulling changesets, we need to take some extra care to get
 # largefiles updated remotely
 def overridepull(orig, ui, repo, source=None, **opts):
@@ -806,9 +801,26 @@ def overridepull(orig, ui, repo, source=None, **opts):
         ui.status(_("%d largefiles cached\n") % numcached)
     return result
 
+def overridepush(orig, ui, repo, *args, **kwargs):
+    """Override push command and store --lfrev parameters in opargs"""
+    lfrevs = kwargs.pop('lfrev', None)
+    if lfrevs:
+        opargs = kwargs.setdefault('opargs', {})
+        opargs['lfrevs'] = scmutil.revrange(repo, lfrevs)
+    return orig(ui, repo, *args, **kwargs)
+
+def exchangepushoperation(orig, *args, **kwargs):
+    """Override pushoperation constructor and store lfrevs parameter"""
+    lfrevs = kwargs.pop('lfrevs', None)
+    pushop = orig(*args, **kwargs)
+    pushop.lfrevs = lfrevs
+    return pushop
+
+revsetpredicate = registrar.revsetpredicate()
+
+@revsetpredicate('pulled()')
 def pulledrevsetsymbol(repo, subset, x):
-    """``pulled()``
-    Changesets that just has been pulled.
+    """Changesets that just has been pulled.
 
     Only available with largefiles from pull --lfrev expressions.
 
@@ -959,22 +971,13 @@ def overridearchive(orig, repo, dest, node, kind, decode=True, matchfn=None,
 
             f = lfutil.splitstandin(f)
 
-            def getdatafn():
-                fd = None
-                try:
-                    fd = open(path, 'rb')
-                    return fd.read()
-                finally:
-                    if fd:
-                        fd.close()
-
-            getdata = getdatafn
+            getdata = lambda: util.readfile(path)
         write(f, 'x' in ff and 0o755 or 0o644, 'l' in ff, getdata)
 
     if subrepos:
         for subpath in sorted(ctx.substate):
             sub = ctx.workingsub(subpath)
-            submatch = match_.narrowmatcher(subpath, matchfn)
+            submatch = match_.subdirmatcher(subpath, matchfn)
             sub._repo.lfstatus = True
             sub.archive(archiver, prefix, submatch)
 
@@ -1016,22 +1019,13 @@ def hgsubrepoarchive(orig, repo, archiver, prefix, match=None):
 
             f = lfutil.splitstandin(f)
 
-            def getdatafn():
-                fd = None
-                try:
-                    fd = open(os.path.join(prefix, path), 'rb')
-                    return fd.read()
-                finally:
-                    if fd:
-                        fd.close()
-
-            getdata = getdatafn
+            getdata = lambda: util.readfile(os.path.join(prefix, path))
 
         write(f, 'x' in ff and 0o755 or 0o644, 'l' in ff, getdata)
 
     for subpath in sorted(ctx.substate):
         sub = ctx.workingsub(subpath)
-        submatch = match_.narrowmatcher(subpath, match)
+        submatch = match_.subdirmatcher(subpath, match)
         sub._repo.lfstatus = True
         sub.archive(archiver, prefix + repo._path + '/', submatch)
 
@@ -1047,6 +1041,13 @@ def overridebailifchanged(orig, repo, *args, **kwargs):
     if s.modified or s.added or s.removed or s.deleted:
         raise error.Abort(_('uncommitted changes'))
 
+def postcommitstatus(orig, repo, *args, **kwargs):
+    repo.lfstatus = True
+    try:
+        return orig(repo, *args, **kwargs)
+    finally:
+        repo.lfstatus = False
+
 def cmdutilforget(orig, ui, repo, match, prefix, explicitonly):
     normalmatcher = composenormalfilematcher(match, repo[None].manifest())
     bad, forgot = orig(ui, repo, normalmatcher, prefix, explicitonly)
@@ -1073,8 +1074,7 @@ def cmdutilforget(orig, ui, repo, match, prefix, explicitonly):
 
     # Need to lock because standin files are deleted then removed from the
     # repository and we could race in-between.
-    wlock = repo.wlock()
-    try:
+    with repo.wlock():
         lfdirstate = lfutil.openlfdirstate(ui, repo)
         for f in forget:
             if lfdirstate[f] == 'a':
@@ -1086,8 +1086,6 @@ def cmdutilforget(orig, ui, repo, match, prefix, explicitonly):
         for f in standins:
             util.unlinkpath(repo.wjoin(f), ignoremissing=True)
         rejected = repo[None].forget(standins)
-    finally:
-        wlock.release()
 
     bad.extend(f for f in rejected if f in m.files())
     forgot.extend(f for f in forget if f not in rejected)
@@ -1246,8 +1244,7 @@ def overridepurge(orig, ui, repo, *dirs, **opts):
     orig(ui, repo, *dirs, **opts)
     repo.status = oldstatus
 def overriderollback(orig, ui, repo, **opts):
-    wlock = repo.wlock()
-    try:
+    with repo.wlock():
         before = repo.dirstate.parents()
         orphans = set(f for f in repo.dirstate
                       if lfutil.isstandin(f) and repo.dirstate[f] != 'r')
@@ -1281,8 +1278,6 @@ def overriderollback(orig, ui, repo, **opts):
         for lfile in orphans:
             lfdirstate.drop(lfile)
         lfdirstate.write()
-    finally:
-        wlock.release()
     return result
 
 def overridetransplant(orig, ui, repo, *revs, **opts):
@@ -1358,10 +1353,12 @@ def overridecat(orig, ui, repo, file1, *pats, **opts):
         err = 0
     return err
 
-def mergeupdate(orig, repo, node, branchmerge, force, partial,
+def mergeupdate(orig, repo, node, branchmerge, force,
                 *args, **kwargs):
-    wlock = repo.wlock()
-    try:
+    matcher = kwargs.get('matcher', None)
+    # note if this is a partial update
+    partial = matcher and not matcher.always()
+    with repo.wlock():
         # branch |       |         |
         #  merge | force | partial | action
         # -------+-------+---------+--------------
@@ -1384,7 +1381,7 @@ def mergeupdate(orig, repo, node, branchmerge, force, partial,
         pctx = repo['.']
         for lfile in unsure + s.modified:
             lfileabs = repo.wvfs.join(lfile)
-            if not os.path.exists(lfileabs):
+            if not repo.wvfs.exists(lfileabs):
                 continue
             lfhash = lfutil.hashrepofile(repo, lfile)
             standin = lfutil.standin(lfile)
@@ -1399,7 +1396,7 @@ def mergeupdate(orig, repo, node, branchmerge, force, partial,
 
         oldstandins = lfutil.getstandinsstate(repo)
 
-        result = orig(repo, node, branchmerge, force, partial, *args, **kwargs)
+        result = orig(repo, node, branchmerge, force, *args, **kwargs)
 
         newstandins = lfutil.getstandinsstate(repo)
         filelist = lfutil.getlfilestoupdate(oldstandins, newstandins)
@@ -1410,8 +1407,6 @@ def mergeupdate(orig, repo, node, branchmerge, force, partial,
                                 normallookup=partial)
 
         return result
-    finally:
-        wlock.release()
 
 def scmutilmarktouched(orig, repo, files, *args, **kwargs):
     result = orig(repo, files, *args, **kwargs)
